# ========================================================================
# TestDigitRecognition.py: Simple program to recognize digits 0 through 9.
# 
# Author: pavankalyan
# Modified by: Mark Austin                                   December 2021 
# ========================================================================

import numpy as np
import matplotlib.pyplot as plt
from sklearn.datasets import load_digits
from sklearn.preprocessing import StandardScaler
import pylab as pl
import random

# ========================================================================
# Useful activation functions and their derivatives ...
# ========================================================================

def sigmoid_(Z):
    return 1/(1+np.exp(-Z))

def relu_(Z):
    return Z*(Z>0)

def drelu_(Z):
    return 1. *(Z>0)

def dsigmoid_(Z):
    return sigmoid_(Z)*(1-sigmoid_(Z))

def sigmoid(Z):
    return sigmoid_(Z),Z

def relu(Z):
    return relu_(Z),Z

def linear_activation_forward(A_prev,W,b,activation):
    if activation == "sigmoid":
        Z, linear_cache = linear_forward(A_prev,W,b)
        A, activation_cache = sigmoid(Z)

    elif activation == "relu":
        Z, linear_cache = linear_forward(A_prev,W,b)
        A, activation_cache = relu(Z)

    assert (A.shape == (W.shape[0], A_prev.shape[1]))
    cache = (linear_cache, activation_cache)

    return A, cache

# ========================================================================
# Initialize parameters for deep neural networks ....
# ========================================================================

def initialize_parameters_deep(layer_dims):
    np.random.seed(3)
    parameters = {}
    L = len(layer_dims)
    for l in range(1, L):
        parameters['W' + str(l)] = np.random.randn(layer_dims[l],layer_dims[l-1])*0.01
        parameters['b' + str(l)] = np.zeros((layer_dims[l],1))

        assert(parameters['W' + str(l)].shape == (layer_dims[l], layer_dims[l-1]))
        assert(parameters['b' + str(l)].shape == (layer_dims[l], 1))
    return parameters

def linear_forward(A, W, b):
    Z = np.dot(W,A)+b
    assert(Z.shape == (W.shape[0], A.shape[1]))
    cache = (A, W, b)

    return Z, cache

# ========================================================================
# Forward propogation for L layer neural network ...
# ========================================================================

def L_model_forward(X, parameters):
    caches = []
    A = X
    L = len(parameters) // 2
    for l in range(1, L):
        A_prev = A
        A, cache = linear_activation_forward(A_prev,parameters['W'+str(l)],parameters['b'+str(l)],"relu")
        caches.append(cache)
    AL, cache = linear_activation_forward(A,parameters['W'+str(L)],parameters['b'+str(L)],"sigmoid")
    caches.append(cache)
    #assert(AL.shape == (1,X.shape[1]))
    return AL, caches

# ========================================================================
# Cost function ...
# ========================================================================

def compute_cost(AL, Y):
    m=Y.shape[1]
    cost = -(1/m)*np.sum((Y*np.log(AL)+(1-Y)*np.log(1-AL)))
    cost=np.squeeze(cost)
    assert(cost.shape == ())
    return cost

def linear_backward(dZ, cache):
    A_prev, W, b = cache
    m = A_prev.shape[1]
    dW = (1/m)*np.dot(dZ,A_prev.T)
    db = (1/m)*np.sum(dZ,axis=1,keepdims=True)
    dA_prev = np.dot(W.T,dZ)

    assert (dA_prev.shape == A_prev.shape)
    assert (dW.shape == W.shape)
    assert (db.shape == b.shape)

    return dA_prev, dW, db

def relu_backward(dA,activation_cache):
    return dA* drelu_(activation_cache)

def sigmoid_backward(dA,activation_cache):
    return dA* dsigmoid_(activation_cache)

def linear_activation_backward(dA, cache, activation):
    linear_cache, activation_cache = cache
    if activation == "relu":
        dZ = relu_backward(dA,activation_cache)
        dA_prev, dW, db = linear_backward(dZ,linear_cache)
    elif activation == "sigmoid":
        dZ = sigmoid_backward(dA,activation_cache)
        dA_prev, dW, db = linear_backward(dZ,linear_cache)
    return dA_prev,dW,db

# ========================================================================
# Back propogation ...
# ========================================================================

def L_model_backward(AL, Y, caches):
    grads = {}
    L = len(caches)
    m = AL.shape[1]
    #Y = Y.reshape(AL.shape)

    dAL = - (np.divide(Y, AL) - np.divide(1 - Y, 1 - AL))

    current_cache = caches[L-1]
    grads["dA" + str(L-1)], grads["dW" + str(L)], grads["db" + str(L)] = linear_activation_backward(dAL,current_cache,"sigmoid")

    for l in reversed(range(L-1)):
        current_cache = caches[l]
        dA_prev_temp, dW_temp, db_temp = linear_activation_backward(grads["dA"+str(l+1)],current_cache,"relu")
        grads["dA" + str(l)] = dA_prev_temp
        grads["dW" + str(l + 1)] = dW_temp
        grads["db" + str(l + 1)] = db_temp
    return grads

# ========================================================================
# Update parameters ...
# ========================================================================

def update_parameters(parameters, grads, learning_rate):
    L = len(parameters) // 2
    for l in range(L):
        parameters["W" + str(l+1)] = parameters["W" + str(l+1)]-(learning_rate)*grads["dW"+str(l+1)]
        parameters["b" + str(l+1)] = parameters["b" + str(l+1)]-(learning_rate)*grads["db"+str(l+1)]
    return parameters

# ========================================================================
# N-Layer Neural Network Model ...
# ========================================================================

def L_layer_model(X, Y, layers_dims, learning_rate = 0.00042, num_iterations = 3000, print_cost=False):
    np.random.seed(1)
    costs = []

    parameters = initialize_parameters_deep(layers_dims)

    for i in range(0, num_iterations):
        AL, caches = L_model_forward(X, parameters)
        cost = compute_cost(AL, Y)
        grads = L_model_backward(AL, Y, caches)
        parameters = update_parameters(parameters, grads, learning_rate)
        if print_cost and i % 1000 == 0:
            print ("Cost after iteration %i: %f" %(i, cost))
        if print_cost and i % 1000 == 0:
            costs.append(cost)

    # plot the cost

    plt.plot(np.squeeze(costs))
    plt.ylabel('cost')
    plt.xlabel('iterations (per tens)')
    plt.title("Learning rate =" + str(learning_rate))
    plt.show()

    return parameters

# ========================================================================
# Predict L layer ...
# ========================================================================

def predict_L_layer(X,parameters):
    AL,caches=L_model_forward(X,parameters)
    prediction=np.argmax(AL,axis=0)
    return prediction.reshape(1,prediction.shape[0])

# ========================================================================
# main method ...
# ========================================================================

def main():
    print("--- Enter TestRecognizeDigits.main()     ... ");
    print("--- ======================================== ... ");

    # Part 1: Load and display digits ...

    print("--- load and display digits ... ");

    digits=load_digits()

    pl.gray()
    pl.matshow(digits.images[0])
    pl.show()

    digits.images[0]
    digits.target[0]

    images_and_labels=list(zip(digits.images,digits.target))
    plt.figure(figsize=(5,5))
    for index,(image,label) in enumerate(images_and_labels[:15]):
       plt.subplot(3,5,index+1)
       plt.axis('off')
       plt.imshow(image,cmap=plt.cm.gray_r,interpolation='nearest')
       plt.title('%i' % label)

    # Part 2: Define variables ...

    print("--- define variables ... ");

    n_samples=len(digits.images)
    print(n_samples)

    x=digits.images.reshape((n_samples,-1))
    print(x.shape)

    y=digits.target
    print(y.shape)

    from sklearn.model_selection import train_test_split

    # Part 3: Split data into train and test sets ...

    print("--- split data into train and test sets ... ");

    X_train, X_test, y_train, y_test = train_test_split(x, y, test_size = 0.2, random_state = 0)

    # Part 4: Feature scaling ...

    print("--- feature scaling ... ");

    sc = StandardScaler()
    X_train = sc.fit_transform(X_train)
    X_test =  sc.transform(X_test)

    X_train=X_train.T
    X_test=X_test.T
    y_train=y_train.reshape(y_train.shape[0],1)
    y_test=y_test.reshape(y_test.shape[0],1)
    y_train=y_train.T
    y_test=y_test.T

    Y_train_=np.zeros((10,y_train.shape[1]))
    for i in range(y_train.shape[1]):
        Y_train_[y_train[0,i],i]=1

    Y_test_=np.zeros((10,y_test.shape[1]))
    for i in range(y_test.shape[1]):
        Y_test_[y_test[0,i],i]=1

    n_x=X_train.shape[0]
    n_h=10
    n_y=Y_train_.shape[0]

    # Part 4: Initialize parameters for deep networks ...

    print("--- Initialize parameters for deep networks ... ");

    parameters = initialize_parameters_deep([n_x,10,n_y])
    W1 = parameters["W1"]
    b1 = parameters["b1"]

    A = X_train
    Z, linear_cache = linear_forward(A, W1, b1)

    # Part 5: Create N-layer neural network ...

    print("--- Create N-layer neural network ... ");

    layers_dims = [n_x,60,10]
    parameters = L_layer_model(X_train, Y_train_, layers_dims, num_iterations = 50000, print_cost = True)

    # Part 6: Compute predicitions ...

    print("--- Compute predictions ... ");

    predictions_train_L = predict_L_layer(X_train, parameters)
    print(np.sum(predictions_train_L==y_train))

    predictions_test_L=predict_L_layer(X_test,parameters)
    print(np.sum(predictions_test_L==y_test))

    for j in range(15):
       i=random.randint(0,n_samples)
       pl.gray()
       pl.matshow(digits.images[i])
       pl.show()
       img=digits.images[i].reshape((64,1)).T
       img = sc.transform(img)
       img=img.T
       predicted_digit=predict_L_layer(img,parameters)
       print('--- Predicted digit is : ' + str(predicted_digit))
       print('--- True digit is: ' + str(y[i]))

    print("--- ======================================== ... ");
    print("--- Leave TestRecognizeDigits.main()     ... ");

# call the main method ...

main()
