# ========================================================================
# TestRecurrentNeuralNetwork.py
# 
# Author: 
# Modified by: Mark Austin                                   December 2021 
# ========================================================================

import numpy as np
import os
from pathlib import Path

import tensorflow as tf
from tensorflow import keras

import sys
import sklearn

import matplotlib as mpl
import matplotlib.pyplot as plt
mpl.rc('axes', labelsize=14)
mpl.rc('xtick', labelsize=12)
mpl.rc('ytick', labelsize=12)

# ========================================================================
# Save images ...
# ========================================================================

PROJECT_ROOT_DIR = "."
CHAPTER_ID = "rnn"
IMAGES_PATH = os.path.join(PROJECT_ROOT_DIR, "images", CHAPTER_ID)
os.makedirs(IMAGES_PATH, exist_ok=True)

def save_fig(fig_id, tight_layout=True, fig_extension="png", resolution=300):
    path = os.path.join(IMAGES_PATH, fig_id + "." + fig_extension)
    print("Saving figure", fig_id)
    if tight_layout:
        plt.tight_layout()
    plt.savefig(path, format=fig_extension, dpi=resolution)

# ========================================================================
# Generate time series ...
# ========================================================================

def generate_time_series(batch_size, n_steps):

    print("--- Enter generate_time_series() ...");

    freq1, freq2, offsets1, offsets2 = np.random.rand(4, batch_size, 1)
    time = np.linspace(0, 1, n_steps)

    print("---   freq1.shape: ", freq1.shape );
    print("---   freq2.shape: ", freq2.shape );
    print("---   offsets1.shape: ", offsets1.shape );
    print("---   offsets2.shape: ", offsets2.shape );

    series  = 0.5 * np.sin((time - offsets1) * (freq1 * 10 + 10)) #   wave 1
    series += 0.2 * np.sin((time - offsets2) * (freq2 * 20 + 20)) # + wave 2
#   series += 0.1 * (np.random.rand(batch_size, n_steps) - 0.5)   # + noise

    print("---   series.shape: ", series.shape );

    print("--- Leave generate_time_series() ...");

    return series[..., np.newaxis].astype(np.float32)

# ========================================================================
# Plot time series ...
# ========================================================================

def plot_series(series, y=None, y_pred=None, x_label="$t$", y_label="$x(t)$"):
    n_steps = 50
    plt.plot(series, ".-")
    if y is not None:
        plt.plot(n_steps, y, "bx", markersize=10)
    if y_pred is not None:
        plt.plot(n_steps, y_pred, "ro")
    plt.grid(True)
    if x_label:
        plt.xlabel(x_label, fontsize=16)
    if y_label:
        plt.ylabel(y_label, fontsize=16, rotation=0)
    plt.hlines(0, 0, 100, linewidth=1)
    plt.axis([0, n_steps + 1, -1, 1])

# ========================================================================
# Plot learning curves ...
# ========================================================================

def plot_learning_curves(loss, val_loss):
    plt.plot(np.arange(len(loss)) + 0.5, loss, "b.-", label="Training loss")
    plt.plot(np.arange(len(val_loss)) + 1, val_loss, "r.-", label="Validation loss")
    plt.gca().xaxis.set_major_locator(mpl.ticker.MaxNLocator(integer=True))
    plt.axis([1, 20, 0, 0.05])
    plt.legend(fontsize=14)
    plt.xlabel("Epochs")
    plt.ylabel("Loss")
    plt.grid(True)

# ========================================================================
# plot multiple forecasts ...
# ========================================================================

def plot_multiple_forecasts(X, Y, Y_pred):
    n_steps = X.shape[1]
    ahead = Y.shape[1]
    plot_series(X[0, :, 0])
    plt.plot(np.arange(n_steps, n_steps + ahead), Y[0, :, 0], "ro-", label="Actual")
    plt.plot(np.arange(n_steps, n_steps + ahead), Y_pred[0, :, 0], "bx-", label="Forecast", markersize=10)
    plt.axis([0, n_steps + ahead, -1, 1])
    plt.legend(fontsize=14)

# ========================================================================
# last time mean square error ...
# ========================================================================

def last_time_step_mse(Y_true, Y_pred):
    return keras.metrics.mean_squared_error(Y_true[:, -1], Y_pred[:, -1])

# ========================================================================
# main method ...
# ========================================================================

def main():
    print("--- Enter TestRecurrentNeuralNetwork01.main() ...");
    print("--- ========================================= ...");

    # Part 1: Make program output stable across runs ...

    print("--- make program output stable across runs ... ");

    np.random.seed(42)
    tf.random.set_seed(42)

    # Part 2: initialize matplotlib parameters ..

    print("--- initialize matplotlib parameters ... ");

    mpl.rc('axes', labelsize=14)
    mpl.rc('xtick', labelsize=12)
    mpl.rc('ytick', labelsize=12)

    # Part 3: generate time series ...

    print("--- generate time series ... ");

    n_steps = 50
    series = generate_time_series(10000, n_steps + 1)
    X_train, y_train = series[:7000, :n_steps], series[:7000, -1]
    X_valid, y_valid = series[7000:9000, :n_steps], series[7000:9000, -1]
    X_test,  y_test  = series[9000:, :n_steps], series[9000:, -1]

    print("---    series.shape: ", series.shape );
    print("---   X_train.shape: ", X_train.shape );
    print("---   y_train.shape: ", y_train.shape );
    print("---   X_valid.shape: ", X_valid.shape );
    print("---   y_valid.shape: ", y_valid.shape );
    print("---    X_test.shape: ", X_test.shape );
    print("---    y_test.shape: ", y_test.shape );

    print("--- Abbreviated details of X_train ... ");

    # print (X_train)

    print("--- Abbreviated details of y_train ... ");

    # print (y_train)

    # Part 4: plot time series ...

    print("--- plot time series ... ");

    fig, axes = plt.subplots(nrows=1, ncols=3, sharey=True, figsize=(12, 4))
    for col in range(3):
        plt.sca(axes[col])
        plot_series(X_valid[col, :, 0], y_valid[col, 0], 
                    y_label=("$x(t)$" if col==0 else None))
    save_fig("time_series_plot")
    plt.show()

    # Part 5: compute baseline (last point in series) ....

    print("--- compute baseline (last point in series) ... ");

    y_pred = X_valid[:, -1]

    print("---   mean square error of prediction: ", 
                 np.mean(keras.losses.mean_squared_error(y_valid, y_pred)));

    # Part 6: create plot of series + baseline point ....

    print("--- create plot of series + baseline point ... ");

    plot_series( X_valid[0, :, 0], y_valid[0, 0], y_pred[0, 0] )
    plt.show()

    # Part 7: compute linear prediction (train on 7000 samples, validate on 2000 samples) ....

    print("--- compute linear prediction ... ");
    print("--- procedure: train: 7000 samples, validate: 2000 samples ... ");

    model = keras.models.Sequential([
       keras.layers.Flatten(input_shape=[50, 1]),
       keras.layers.Dense(1)
    ])

    model.compile(loss="mse", optimizer="adam")
    history = model.fit(X_train, y_train, epochs=20,
                        validation_data=(X_valid, y_valid))

    print("--- model evaluation: ", model.evaluate(X_valid, y_valid));

    # Part 8: plot learning curves ...

    print("--- plot learning curves ... ");

    plot_learning_curves( history.history["loss"],
                          history.history["val_loss"])
    plt.show()

    # Part 9: plot series + linear prediction point ...

    print("--- plot series + linear prediction point ... ");

    y_pred = model.predict(X_valid)
    plot_series(X_valid[0, :, 0], y_valid[0, 0], y_pred[0, 0])
    plt.show()

    # ==========================================================
    # Part 10: Use a Simple RNN ...
    # ==========================================================

    print("");
    print("--- ================================== ... ");
    print("--- Use a Simple RNN ... ");
    print("--- ================================== ... ");
    print("");

    np.random.seed(42)
    tf.random.set_seed(42)

    model = keras.models.Sequential([
        keras.layers.SimpleRNN(1, input_shape=[None, 1])
    ])

    optimizer = keras.optimizers.Adam(lr=0.005)

    print("--- print model summary ... ");

    print( model.summary() )

    print("--- print model layers ... ");

    for layer in model.layers:
        print(layer.name, layer.inbound_nodes, layer.outbound_nodes)

    print("--- compile and train model ... ");

    model.compile(loss="mse", optimizer=optimizer)
    history = model.fit(X_train, y_train, epochs=20,
                        validation_data=(X_valid, y_valid))

    print("--- model evaluation: ", model.evaluate(X_valid, y_valid));
    print("--- plot learning curves ... ");

    plot_learning_curves( history.history["loss"],
                          history.history["val_loss"])
    plt.show()

    print("--- plot series + simple rnn prediction linear prediction point ... ");

    y_pred = model.predict(X_valid)
    plot_series(X_valid[0, :, 0], y_valid[0, 0], y_pred[0, 0])
    plt.show()

    # ==========================================================
    # Part 11: Use a Deep RNN ...
    # ==========================================================

    print("");
    print("--- ================================== ... ");
    print("--- Use a Deep RNN ... ");
    print("--- ================================== ... ");
    print("");

    np.random.seed(42)
    tf.random.set_seed(42)

    model = keras.models.Sequential([
        keras.layers.SimpleRNN(20, return_sequences=True, input_shape=[None, 1]),
        keras.layers.SimpleRNN(20, return_sequences=True),
        keras.layers.SimpleRNN(1)
    ])

    print("--- print model summary ... ");

    print( model.summary() )

    print("--- print model layers ... ");

    for layer in model.layers:
        print(layer.name, layer.inbound_nodes, layer.outbound_nodes)

    print("--- compile and train model ... ");

    model.compile(loss="mse", optimizer="adam")
    history = model.fit(X_train, y_train, epochs=20,
                        validation_data=(X_valid, y_valid))

    print("--- model evaluation: ", model.evaluate(X_valid, y_valid));
    print("--- plot learning curves ... ");

    plot_learning_curves( history.history["loss"],
                          history.history["val_loss"])
    plt.show()

    print("--- plot series + simple rnn prediction linear prediction point ... ");

    y_pred = model.predict(X_valid)
    plot_series(X_valid[0, :, 0], y_valid[0, 0], y_pred[0, 0])
    plt.show()

    # ==========================================================
    # Part 12: Forecasting several steps ahead ...
    # ==========================================================

    print("");
    print("--- =============================== ... ");
    print("--- Forecasting several steps ahead ... ");
    print("--- =============================== ... ");
    print("");

    np.random.seed(43)

    series = generate_time_series(1, n_steps + 10)
    X_new, Y_new = series[:, :n_steps], series[:, n_steps:]
    X = X_new
    for step_ahead in range(10):
       y_pred_one = model.predict(X[:, step_ahead:])[:, np.newaxis, :]
       X = np.concatenate([X, y_pred_one], axis=1)

    Y_pred = X[:, n_steps:]

    print("---    Y_pred shape: ", Y_pred.shape );
    print("--- plot multiple forecasts ... ");

    plot_multiple_forecasts(X_new, Y_new, Y_pred)
    save_fig("forecast_ahead_plot")
    plt.show()

    print("--- use model to predict next ten values (sequences have nine more values) ... ");

    np.random.seed(42)

    n_steps = 50
    series = generate_time_series(10000, n_steps + 10)
    X_train, Y_train = series[:7000, :n_steps], series[:7000, -10:, 0]
    X_valid, Y_valid = series[7000:9000, :n_steps], series[7000:9000, -10:, 0]
    X_test, Y_test   = series[9000:, :n_steps], series[9000:, -10:, 0]

    print("--- predict next ten values one by one ... ");

    X = X_valid
    for step_ahead in range(10):
       y_pred_one = model.predict(X)[:, np.newaxis, :]
       X = np.concatenate([X, y_pred_one], axis=1)

    Y_pred = X[:, n_steps:, 0]
    print("---   Y_pred shape: ", Y_pred.shape );
    print("---   mean square error of prediction: ", 
                 np.mean(keras.losses.mean_squared_error(Y_valid, Y_pred)));

    # =============================================================
    # Part 14: Create RNN that predicts ten next values at once ...
    # =============================================================

    print("");
    print("--- ========================================================= ... ");
    print("--- Part 14: Create RNN that predicts ten next values at once ... ");
    print("--- ========================================================= ... ");
    print("");

    np.random.seed(42)
    tf.random.set_seed(42)

    print("--- create model ... ");

    model = keras.models.Sequential([
        keras.layers.SimpleRNN(20, return_sequences=True, input_shape=[None, 1]),
        keras.layers.SimpleRNN(20),
        keras.layers.Dense(10)
    ])

    print("--- compile and fit model ... ");
    print("---     X_train: ", X_train.shape );
    print("---     Y_train: ", Y_train.shape );

    model.compile(loss="mse", optimizer="adam")
    history = model.fit(X_train, Y_train, epochs=20, validation_data=(X_valid, Y_valid))

    np.random.seed(43)
    series = generate_time_series(1, 50 + 10)

    X_new, Y_new = series[:, :50, :], series[:, -10:, :]
    Y_pred = model.predict(X_new)[..., np.newaxis]

    print("---     X_new shape: ", X_new.shape );
    print("---     Y_new shape: ", Y_new.shape );
    print("---    Y_pred shape: ", Y_pred.shape );

    print("--- plot multiple forecasts ... ");

    plot_multiple_forecasts(X_new, Y_new, Y_pred)
    plt.show()

    # =============================================================
    # Part 15: Create RNN that progressively forecasts next ten steps
    # at each time step: i.e., 1-11, 2-12 ... thru ... 50-59.
    # 
    # Note: This model is causal. When it makes a prediction it can 
    #       only see past steps.
    # =============================================================

    print("");
    print("--- ================================================================ ... ");
    print("--- Part 15: Create RNN that progressively forecasts next ten steps `... ");
    print("--- ================================================================ ... ");
    print("");

    np.random.seed(42)
    tf.random.set_seed(42)

    n_steps = 50
    series = generate_time_series(10000, n_steps + 10)

    X_train = series[:7000, :n_steps]
    X_valid = series[7000:9000, :n_steps]
    X_test  = series[9000:, :n_steps]
    Y       = np.empty((10000, n_steps, 10))

    for step_ahead in range(1, 10 + 1):
        Y[..., step_ahead - 1] = series[..., step_ahead:step_ahead + n_steps, 0]

    Y_train = Y[:7000]
    Y_valid = Y[7000:9000]
    Y_test  = Y[9000:]

    print("---    X_train shape: ", X_train.shape );
    print("---    Y_train shape: ", Y_train.shape );

    np.random.seed(42)
    tf.random.set_seed(42)

    model = keras.models.Sequential([
       keras.layers.SimpleRNN(20, return_sequences=True, input_shape=[None, 1]),
       keras.layers.SimpleRNN(20, return_sequences=True),
       keras.layers.TimeDistributed(keras.layers.Dense(10))
    ])

    model.compile(loss="mse", optimizer=keras.optimizers.Adam(lr=0.01), metrics=[last_time_step_mse])
    history = model.fit(X_train, Y_train, epochs=20,
                        validation_data=(X_valid, Y_valid))

    np.random.seed(43)
    series = generate_time_series(1, 50 + 10)
    X_new, Y_new = series[:, :50, :], series[:, 50:, :]
    Y_pred = model.predict(X_new)[:, -1][..., np.newaxis]

    print("--- plot multiple forecasts ... ");

    plot_multiple_forecasts(X_new, Y_new, Y_pred)
    plt.show()

    print("--- ======================================== ... ");
    print("--- Leave TestRecurrentNeuralNetwork01.main() ...");

# call the main method ...

main()
